unit SpotLight1;

{ SpotLight demonstration (c)2001 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr

  from Martin Beaudet code on www.developpez.com.

}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WinGL, DelphiGL, AppEvnts, ExtCtrls, ComCtrls, StdCtrls;

type
  TForm1 = class(TForm)
    WinGL1: TWinGL;
    GLLight1: TGLLight;
    ApplicationEvents1: TApplicationEvents;
    Timer1: TTimer;
    Panel1: TPanel;
    TrackBar1: TTrackBar;
    Label1: TLabel;
    Panel2: TPanel;
    UseList: TCheckBox;
    Label2: TLabel;
    Label3: TLabel;
    GLTex1: TGLTex;
    procedure WinGL1Paint(Sender: TObject);
    procedure WinGL1CreateGL(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure Timer1Timer(Sender: TObject);
    procedure UseListClick(Sender: TObject);
  private
   DegreRot1, DegreRot2, DegreRot3:single; // litgh source movement
   Frames:integer; // FPS
   Face:integer;   // Face number when using CallList
   last:integer;   // last GetTickCount
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

var
 tcount:integer; // number of triangle drawn

// subdivise or draw a triangle
Procedure SubTriangle(x1,y1,x2,y2,x3,y3, l1,l2,l3:single; iter:integer);
 const
  min=3;
 var
  l:integer;
  xm,ym,lm:single;
 begin
  if iter=0 then begin // last iteration, draw
   inc(tcount);
   glVertex2f(x1,y1);
   glVertex2f(x2,y2);
   glVertex2f(x3,y3);
  end else begin
   if (l1>l2)and(l1>l3) then begin // check longuest edge
    l:=1;
   end else begin
    if l2>l3 then l:=2 else l:=3;
   end;
   dec(iter);
   case l of
    1: begin // first edge
        if l1<min then iter:=0; // last iter for small edge
        xm:=x1+(x2-x1)/2; // middle of the edge
        ym:=y1+(y2-y1)/2;
        l1:=l1/2; // half size if the edge
        lm:=sqrt(sqr(xm-x3)+sqr(ym-y3));  // length of the new edge
        SubTriangle(x1,y1,xm,ym,x3,y3, l1,lm,l3, iter); // subdivise in two triangles
        SubTriangle(xm,ym,x2,y2,x3,y3, l1,l2,lm, iter);
       end;
    2: begin // 2nd edge is longuer
        if l2<min then iter:=0;
        xm:=x2+(x3-x2)/2;
        ym:=y2+(y3-y2)/2;
        l2:=l2/2;
        lm:=sqrt(sqr(xm-x1)+sqr(ym-y1));
        SubTriangle(x1,y1,x2,y2,xm,ym, l1,l2,lm, iter);
        SubTriangle(x1,y1,xm,ym,x3,y3, lm,l2,l3, iter);
       end;
    3: begin // 3rd is longuer
        if l3<min then iter:=0;
        xm:=x3+(x1-x3)/2;
        ym:=y3+(y1-y3)/2;
        l3:=l3/2;
        lm:=sqrt(sqr(xm-x2)+sqr(ym-y2));
        SubTriangle(x1,y1,x2,y2,xm,ym, l1,lm,l3, iter);
        SubTriangle(x2,y2,x3,y3,xm,ym, l2,l3,lm, iter);
       end;
   end;
  end;
 end;

Procedure Triangle(x1,y1, x2,y2, x3,y3:single);
 var
  l1,l2,l3:single;
 begin
  l1:=sqrt(sqr(x2-x1)+sqr(y2-y1)); // edge lengths
  l2:=sqrt(sqr(x3-x2)+sqr(y3-y2));
  l3:=sqrt(sqr(x1-x3)+sqr(y1-y3));
  SubTriangle(x1,y1,x2,y2,x3,y3,l1,l2,l3, Form1.TrackBar1.Position);
 end;

procedure TForm1.WinGL1CreateGL(Sender: TObject);
begin
  glTex1.GenTexture; // build GL texture
  glMaterialf(GL_FRONT, GL_SHININESS, 20);
  Face:=0; // no list
end;

procedure TForm1.WinGL1Paint(Sender: TObject);
var
 p0,p1:array[0..3] of single;
 r:single;
 t:integer;
begin
 inc(Frames);
 if WinGL1.GLMode=glManual then begin // not the case :)
  glClear(GL_COLOR_BUFFER_BIT Or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
 end;

 gluLookAt(
  0.0,-90.0, 70.0,
  0.0,-10.0,  0.0,
  0.0,  1.0,  0.0
 );

 { move source ligths }
 p0[0]:=-4 * Cos(DegreRot1) *-5 * Cos(DegreRot2);
 p0[1]:= 5 * Sin(DegreRot1);
 p0[2]:=-6 * Cos(DegreRot1) * 4 * Sin(DegreRot2);
 p0[3]:=1;
 glLight1.SpotExponent:=30+(10*Sin(DegreRot2)); // use OpenGL Lighting
 glLight1.MoveTo(p0[0],p0[1],p0[2],p0[3]);

 p1[0]:=-3 * Cos(DegreRot3) * 6 * Cos(DegreRot1);
 p1[1]:=-6 * Sin(DegreRot1);
 p1[2]:= 5 * Cos(DegreRot1) * 5 * Sin(DegreRot2);
 p1[3]:=1;

 { the face}
  glPushMatrix();
   glTranslated( 0.0, -25.0, -25.0 );
   glRotated( -35, 1.0, 0.0, 0.0 );
   glColorMaterial( GL_FRONT, GL_AMBIENT ); glColor4f( 0.5, 0.5, 0.6, 1.0 );
   glColorMaterial( GL_FRONT, GL_DIFFUSE ); glColor4f( 0.5, 0.2, 0.8, 1.0 );
   glColorMaterial( GL_FRONT, GL_SPECULAR); glColor4f( 0.9, 0.8, 1.0, 1.0 );
  { fill the StencilBuffer with 1 where we draw the face }
   glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
   glStencilFunc(GL_ALWAYS,1,1);
  // set normal vector
   glNormal3f( 0.0, 0.0, 1.0 );
  // use a calllist ?
   if (Face=0) then begin // we don't have one
    if (UseList.Checked) then begin // need to create it
     Face:=glGenLists(1);
     glNewList(Face,GL_COMPILE_AND_EXECUTE);
     panel2.caption:=' new list !';
    end;
    tcount:=0;
    glBegin(GL_TRIANGLES); // build list or just draw
     Triangle(-30,-30, -30,+30, +30,+30);
     Triangle(-30,-30, +30,+30, +30,-30);
    glEnd;
    if (UseList.Checked) then glEndList else panel2.Caption:=IntToStr(tcount)+' faces';
   end else begin
    glCallList(Face); // call the builded list
    panel2.caption:=' list';
   end;

  // texture based lighting
   glDisable(GL_LIGHTING);   // no light
   glDisable(GL_DEPTH_TEST); // need to override previous face, no depth test
   glEnable(GL_TEXTURE_2D);  // need a texture
   glEnable(GL_BLEND); // need color blending

   glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);
   glTex1.Bind; // bind it
   glColor4f(0.2,0.2,0.1,0.5); // spot color
   glStencilFunc(GL_EQUAL,1,1); // draw only where StencilBuffer=1, where we've put the face
   glBegin(GL_QUADS);
    r:=p1[2]/2+15;   // radius
    glTexCoord2i(0,0); glVertex2f(p1[0]-r,p1[1]-r);
    glTexCoord2i(0,1); glVertex2f(p1[0]-r,p1[1]+r);
    glTexCoord2i(1,1); glVertex2f(p1[0]+r,p1[1]+r);
    glTexCoord2i(1,0); glVertex2f(p1[0]+r,p1[1]-r);
   glEnd;
   glStencilFunc(GL_ALWAYS,0,0); // no more stencil test
   glColor4f(1,1,1,1); // restore color
   glDisable(GL_BLEND); // remove blending
   glDisable(GL_TEXTURE_2D); // no more texture
   glEnable(GL_DEPTH_TEST); // ready for next frame
//   glEnable( GL_LIGHTING );

  glPopMatrix();

//   glDisable( GL_LIGHTING );
      glPointSize( 3.0 );
      glColor3f( 1.0, 1.0, 1.0);
      glBegin( GL_POINTS );
        glVertex3d( p0[0], p0[1], p0[2] );
        glVertex3d( p1[0], p1[1], p1[2] );
      glEnd();
   glEnable( GL_LIGHTING );

   t:=last;
   last:=GetTickCount;
   t:=last-t;
   DegreRot1 := (DegreRot1 + 0.05*t/20);
   DegreRot2 := (DegreRot2 + 0.09*t/20);
   DegreRot3 := (DegreRot3 + 0.03*t/20);
   TrackBar1.SelEnd:=TrackBar1.Position;
end;

procedure TForm1.ApplicationEvents1Idle(Sender: TObject;
  var Done: Boolean);
begin
 Done:=False;
 WinGL1.Invalidate;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
 Caption:=IntToStr(Frames)+' FPS';
 Frames:=0;
end;

procedure TForm1.UseListClick(Sender: TObject);
begin
 if Face<>0 then begin
  glDeleteLists(Face,1);
  glTex1.Release;
  Face:=0;
 end;
end;

end.
